/* setinfinity.c --- get/set Rocket flags */
#include <stdlib.h>	
#include <stdio.h>
#include <fcntl.h>
#include <termios.h>
#include <string.h>
#include <errno.h>
#include <grp.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <unistd.h>

#define MKDEV(a,b) ((int)((((a) & 0xff) << 8) | ((b) & 0xff)))
//-----------------------------------------------------------------------------------
#define ROCKET_CALLOUT_NOHUP    0x00000001
#define ROCKET_FORCE_CD         0x00000002
#define ROCKET_HUP_NOTIFY       0x00000004
#define ROCKET_SPLIT_TERMIOS    0x00000008
#define ROCKET_SPD_MASK         0x00000070
#define ROCKET_SPD_HI           0x00000010      /* Use 56000 instead of 38400 bps */
#define ROCKET_SPD_VHI          0x00000020      /* Use 115200 instead of 38400 bps */
#define ROCKET_SPD_SHI          0x00000030      /* Use 230400 instead of 38400 bps */
#define ROCKET_SPD_WARP         0x00000040      /* Use 460800 instead of 38400 bps */
#define ROCKET_SPD_WARP13       0x00000050      /* Use 921600 instead of 38400 bps */
#define ROCKET_SAK              0x00000080
#define ROCKET_SESSION_LOCKOUT  0x00000100
#define ROCKET_PGRP_LOCKOUT     0x00000200
#define ROCKET_RTS_TOGGLE       0x00000400
#define ROCKET_IS_SMPTE         0x00000800       /* Is an SMPTE card */
#define ROCKET_MODE_MASK        0x00007000
#define ROCKET_MODE_RS232       0x00000000
#define ROCKET_MODE_RS422       0x00001000
#define ROCKET_MODE_RS485       0x00002000       /* 485 2-wire */
#define ROCKET_MODE_RS485FDS    0x00003000       /* 485 4-wire slave */
#define ROCKET_MODE_RS485FDM    0x00004000       /* 485 4-wire master */
#define ROCKET_FLAGS            0x00007FFF

#define ROCKET_INITIALIZED      0x80000000      /* Port is active */
#define ROCKET_CLOSING          0x40000000      /* Serial port is closing */
#define ROCKET_NORMAL_ACTIVE    0x20000000      /* Normal port is active */

#define ROCKET_USR_MASK 0x0071  /* Legal flags that non-privileged
                                 * users can set or reset */
/*
 * For closing_wait and closing_wait2
 */
#define ROCKET_CLOSING_WAIT_NONE        65535
#define ROCKET_CLOSING_WAIT_INF         0

#define RCKP_GET_STRUCT               0x00525001
#define RCKP_GET_CONFIG         0x00525002
#define RCKP_SET_CONFIG         0x00525003
#define RCKP_GET_PORTS          0x00525004
#define RCKP_GET_VERSION        0x00525006
//#define RCKP_RESET_USE_COUNT     0x00525007
#define RP2_RESET_USE_COUNT     0x00525007	//BKA

typedef struct {
        unsigned long model;
        char modelString[80];
        unsigned long numPorts;
        int startingPortNumber;
} rocketModel_t;

struct rocket_ports {
        int tty_major;
        int callout_major;
        rocketModel_t rocketModel[8];
};

struct rocket_config {
        int line;
        int flags;
        int closing_wait;
        int close_delay;
        int port;
        int reserved[32];
};

//-----------------------------------------------------------------------------------

#define VERSION_STR "1.06"

char *progname;

int	verbosity = 1;		/* 1 = normal, 0=boot-time, 2=everything */
int	verbose_flag = 0;	/* print results after setting a port */
int	quiet_flag = 0;
int	old_device_name = 0;

#define CMD_FLAG	1
#define CMD_DELAY	2
#define CMD_WAIT	3
#define CMD_COUNTRY	4

#define FLAG_CAN_INVERT	0x0001
#define FLAG_NEED_ARG	0x0002
#define FLAG_READ_ONLY  0x0004

struct flag_type_table {
    int	cmd;
    char	*name;
    int	bits;
    int	mask;
    int	level;
    int	flags;
} flag_type_tbl[] = {
    CMD_FLAG,	"rtsToggle",	ROCKET_RTS_TOGGLE, ROCKET_RTS_TOGGLE, 0, FLAG_CAN_INVERT,
    CMD_FLAG,	"spd_normal",	0,		ROCKET_SPD_MASK,	0, 0,
    CMD_FLAG,	"spd_hi",	ROCKET_SPD_HI, 	ROCKET_SPD_MASK,        0, 0,
    CMD_FLAG,	"spd_vhi",	ROCKET_SPD_VHI,	ROCKET_SPD_MASK,	0, 0,
    CMD_FLAG,	"spd_shi",	ROCKET_SPD_SHI,	ROCKET_SPD_MASK,	0, 0,
    CMD_FLAG,	"spd_warp",	ROCKET_SPD_WARP, ROCKET_SPD_MASK,	0, 0,
    CMD_FLAG,	"spd_warp13",	ROCKET_SPD_WARP13, ROCKET_SPD_MASK,	0, 0,
    CMD_FLAG,   "rs232",        ROCKET_MODE_RS232, ROCKET_MODE_MASK,    0,  0,
    CMD_FLAG,   "rs422",        ROCKET_MODE_RS422, ROCKET_MODE_MASK,    0,  0,
    CMD_FLAG,   "rs485_2-wire",         ROCKET_MODE_RS485,    ROCKET_MODE_MASK,    0,  0,
    CMD_FLAG,   "rs485_4-wire_slave",   ROCKET_MODE_RS485FDS, ROCKET_MODE_MASK,    0,  0,
    CMD_FLAG,   "rs485_4-wire_master",  ROCKET_MODE_RS485FDM, ROCKET_MODE_MASK,    0,  0,
    CMD_FLAG,   "SMPTE",        ROCKET_IS_SMPTE,   ROCKET_IS_SMPTE,     0,  FLAG_READ_ONLY,
    CMD_FLAG, 	"SAK", 		ROCKET_SAK, 	ROCKET_SAK, 	0, FLAG_CAN_INVERT,
    CMD_FLAG,	"hup_notify",	ROCKET_HUP_NOTIFY, ROCKET_HUP_NOTIFY, 0, FLAG_CAN_INVERT,
    CMD_FLAG,	"split_termios", ROCKET_SPLIT_TERMIOS, ROCKET_SPLIT_TERMIOS, 2, FLAG_CAN_INVERT,
    CMD_FLAG,	"session_lockout", ROCKET_SESSION_LOCKOUT, ROCKET_SESSION_LOCKOUT, 2, FLAG_CAN_INVERT,
    CMD_FLAG,	"pgrp_lockout", ROCKET_PGRP_LOCKOUT, ROCKET_PGRP_LOCKOUT, 2, FLAG_CAN_INVERT,
    CMD_FLAG,	"callout_nohup", ROCKET_CALLOUT_NOHUP, ROCKET_CALLOUT_NOHUP, 2, FLAG_CAN_INVERT,

    CMD_DELAY,	"close_delay",	0,		0,		0, FLAG_NEED_ARG,
    CMD_WAIT,	"closing_wait",	0,		0,		0, FLAG_NEED_ARG,
    CMD_COUNTRY,	"country",	0,		0,		0, FLAG_NEED_ARG,
    0,		0,		0,		0,		0, 0,
};
	
 	
int atonum(char *s)
{
    int n;

    while (*s == ' ')
	s++;
    if (strncmp(s, "0x", 2) == 0 || strncmp(s, "0X", 2) == 0)
	sscanf(s + 2, "%x", &n);
    else if (s[0] == '0' && s[1])
	sscanf(s + 1, "%o", &n);
    else
	sscanf(s, "%d", &n);
    return n;
}

void print_flags(struct rocket_config *serinfo,
		 char *prefix, char *postfix)
{
    struct	flag_type_table	*p;
    int	flags;
    int	first = 1;

    flags = serinfo->flags;
	
    for (p = flag_type_tbl; p->name; p++) {
	if (p->cmd != CMD_FLAG)
	    continue;
	if (verbosity < p->level)
	    continue;
	if ((flags & p->mask) == p->bits) {
	    if (first) {
		printf("%s", prefix);
		first = 0;
	    } else
		printf(" ");
	    printf("%s", p->name);
	}
    }
	
    if (!first)
	printf("%s", postfix);
}

/*
 * Set the termios to something sane
 */
int initialize_termios(int fd)
{
    struct termios	t;

    if (tcgetattr(fd, &t) < 0)
	return errno;

    cfsetospeed(&t, B9600);
    cfsetispeed(&t, B9600);
    t.c_iflag &= ~(ISTRIP|INLCR|ICRNL|IGNCR);
    t.c_iflag |= ICRNL;
    t.c_oflag &= ~(OPOST|ONLCR|OCRNL|ONLRET|OFILL);
    t.c_cflag |= CREAD|CLOCAL;
    t.c_lflag &= ~(ISIG|ICANON|ECHO|IEXTEN);

    if (tcsetattr(fd, TCSANOW, &t) < 0)
	return errno;

    return 0;
}


void get_rocket(char *device)
{
    struct rocket_config config;
    int	fd;
    char	buf1[40];

    if ((fd = open(device, O_RDWR|O_NONBLOCK)) < 0) {
	perror(device);
	return;
    }
    if (ioctl(fd, RCKP_GET_CONFIG, &config) < 0) {
	perror("Cannot get rocket config structure");
	close(fd);
	return;
    }
    if (verbosity==2) {
	printf("%s, Line %d, Base port: 0x%.4x\n",
	       device, config.line, config.port);
	if (config.closing_wait == ROCKET_CLOSING_WAIT_INF)
	    strcpy(buf1, "infinte");
	else if (config.closing_wait == ROCKET_CLOSING_WAIT_NONE)
	    strcpy(buf1, "none");
	else
	    sprintf(buf1, "%d", config.closing_wait);
	printf("\tclose_delay: %d, closing_wait: %s\n",
	       config.close_delay, buf1);
	print_flags(&config, "\tFlags: ", "");
	printf("\n\n");
    } else {
	printf("%s, ", device);
	print_flags(&config, "Flags: ", "");
	printf("\n");
    }
    close(fd);
}


void set_rocket(char *device, char ** arg)
{
    struct rocket_config old_config, new_config;
    struct	flag_type_table	*p;
    int	fd;
    int	do_invert = 0;
    char	*word;
	

    if ((fd = open(device, O_RDWR|O_NONBLOCK)) < 0) {
	if (verbosity==0 && errno==ENOENT)
	    exit(201);
	perror(device);
	exit(201);
    }
    if (fcntl(fd, F_SETFL, 0) < 0) {
	perror("fcntl");
	exit(1);
    }
    if (ioctl(fd, RCKP_GET_CONFIG, &old_config) < 0) {
	perror("Cannot get rocket info");
/*		exit(1); */
    }
    new_config = old_config;
    while (*arg) {
	do_invert = 0;
	word = *arg++;
	if (*word == '^') {
	    do_invert++;
	    word++;
	}
	for (p = flag_type_tbl; p->name; p++) {
	    if (!strcasecmp(p->name, word))
		break;
	}
	if (!p->name) {
	    fprintf(stderr, "Invalid flag: %s\n", word);
	    exit(1);
	}
        if (p->flags & FLAG_READ_ONLY) {
            fprintf(stderr, "This flag is read only: %s\n", word);
            exit(1);
        }
	if (do_invert && !(p->flags & FLAG_CAN_INVERT)) {
	    fprintf(stderr, "This flag can not be inverted: %s\n", word);
	    exit(1);
	}
	if ((p->flags & FLAG_NEED_ARG) && !*arg) {
	    fprintf(stderr, "Missing argument for %s\n", word);
	    exit(1);
	}
        if ((p->mask == ROCKET_MODE_MASK) && (new_config.flags & ROCKET_IS_SMPTE)) {
            fprintf(stderr, "Port interface hardware of SMPTE cards can not be changed: %s\n", word);
	    exit(1);
        }
	switch (p->cmd) {
	case CMD_FLAG:
	    new_config.flags &= ~p->mask;
	    if (!do_invert)
	    {
		new_config.flags |= p->bits;
            }
	    break;
	case CMD_DELAY:
	    new_config.close_delay = atonum(*arg++);
	    break;
	case CMD_WAIT:
	    if (!strcasecmp(*arg, "infinite"))
		new_config.closing_wait = ROCKET_CLOSING_WAIT_INF;
	    else if (!strcasecmp(*arg, "none"))
		new_config.closing_wait = ROCKET_CLOSING_WAIT_NONE;
	    else
		new_config.closing_wait = atonum(*arg);
	    arg++;
	    break;
	default:
	    fprintf(stderr, "Internal error: unhandled cmd #%d\n", p->cmd);
	    exit(1);
	}
    }
    if (ioctl(fd, RCKP_SET_CONFIG, &new_config) < 0) {
	perror("Cannot set rocket info");
	exit(1);
    }
    close(fd);
    if (verbose_flag)
	get_rocket(device);
}

void make_device(char *pathname, int mode, int major, int minor,
		 gid_t tty_group)
{
    struct stat statbuf;
    dev_t	device;

    device = MKDEV(major, minor);
    if (lstat(pathname, &statbuf) == 0) {
	if (device == statbuf.st_rdev)
	    return;
	unlink(pathname);
    }
    if (mknod(pathname, mode, device) < 0) {
	perror(pathname);
    }
    chown(pathname, 0, tty_group);
}

void mkdev_rocket(char *device)
{
    struct rocket_ports rocketInfo;
    int	fd, board, port, line;
    gid_t tty_group = 0;
    struct group *grp;
    char pathname[64];
    mode_t mode;

    grp = getgrnam("tty");
    if (grp)
	tty_group = grp->gr_gid;

    if ((fd = open(device, O_RDWR|O_NONBLOCK)) < 0) {
	perror(device);
	return;
    }
    if (ioctl(fd, RCKP_GET_PORTS, &rocketInfo) < 0) {
	perror("Cannot get rocket ports structure");
	close(fd);
	return;
    }

    close(fd);
    mode = S_IFCHR | 0620;
    umask(0);
    fflush(stdout);

    // Walk through the model info struct a board at a time, creating
    // the /dev/ttyRPx devices

    for (board = 0; board < 4; board++) {
	if (!rocketInfo.rocketModel[board].model)
	    break;

	printf("Installing %s, creating /dev/ttyRP%d - %lu\n", 
	       rocketInfo.rocketModel[board].modelString,
	       rocketInfo.rocketModel[board].startingPortNumber,
	       rocketInfo.rocketModel[board].startingPortNumber + rocketInfo.rocketModel[board].numPorts - 1);

	for (line = rocketInfo.rocketModel[board].startingPortNumber;
	     line < rocketInfo.rocketModel[board].startingPortNumber + rocketInfo.rocketModel[board].numPorts;
		line++)
	{
	    //  Create the /dev/ttyRP0 device
	    sprintf(pathname, "/dev/ttyRP%d", line);
	    //printf("DEBUG: Creating %s \n", pathname);
	    make_device(pathname, mode, rocketInfo.tty_major,
			line, tty_group);
    	}
    }
}

void usage()
{
    fprintf(stderr, "setinfinity Version %s\n\n", VERSION_STR);
    fprintf(stderr,
	    "usage: %s rocket-device [cmd1 [arg]] ... \n\n", progname);
    fprintf(stderr, "Available commands: (* = Takes an argument)\n");
    fprintf(stderr, "\t\t(^ = can be preceded by a '^' to turn off the option)\n");
    fprintf(stderr, "\t* close_delay\tset the amount of time (in 1/100 of a\n");
    fprintf(stderr, "\t\t\t\tsecond) that DTR should be kept low\n");
    fprintf(stderr, "\t\t\t\twhile being closed\n");
    fprintf(stderr, "\t* closing_wait\tset the amount of time (in 1/100 of a\n");
    fprintf(stderr, "\t\t\t\tsecond) that the rocket port should wait for\n");
    fprintf(stderr, "\t\t\t\tdata to be drained while being closed\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "\t^ rtsToggle\tEnable the RTS toggle flag on the port\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "\t  rs232\t\tSet port interface hardware to RS-232\n");
    fprintf(stderr, "\t  rs422\t\tSet port interface hardware to RS-422\n");
    fprintf(stderr, "\t  rs485_2-wire\t\tSet port interface hardware to RS-485 2-wire\n");
    fprintf(stderr, "\t  rs485_4-wire_slave\t\tSet port interface hardware to RS-485 4-wire Slave\n");
    fprintf(stderr, "\t  rs485_4-wire_master\t\tSet port interface hardware to RS-485 4-wire Master\n");
    fprintf(stderr, "\n");
#if 0
    fprintf(stderr, "\t^ sak\t\tset the break key as the Secure Attention Key\n");
#endif
    fprintf(stderr, "\t  spd_hi\tuse 56kb instead of 38.4kb\n");
    fprintf(stderr, "\t  spd_vhi\tuse 115kb instead of 38.4kb\n");
    fprintf(stderr, "\t  spd_shi\tuse 230kb instead of 38.4kb\n");
    fprintf(stderr, "\t  spd_warp\tuse 460kb instead of 38.4kb\n");
    fprintf(stderr, "\t  spd_warp13\tuse 921kb instead of 38.4kb\n");//BKA
    fprintf(stderr, "\t  spd_normal\tuse 38.4kb when a baud rate of 38.4kb is selected\n");
    fprintf(stderr, "\n");
    exit(1);
}

main(int argc, char **argv)
{
    int	get_flag = 0;
    int	mkdev_flag = 0;
    int	c;
    extern int optind;
    extern char *optarg;
	
    progname = argv[0];
    if (argc == 1)
	usage();
    while ((c = getopt(argc, argv, "abgoqvMVW")) != EOF) {
	switch (c) {
	case 'a':
	    verbosity = 2;
	    break;
	case 'b':
	    verbosity = 0;
	    break;
	case 'q':
	    quiet_flag++;
	    break;
	case 'v':
	    verbose_flag++;
	    break;
	case 'g':
	    get_flag++;
	    mkdev_flag = 0;
	    break;
	case 'o':
	    old_device_name++;
	    break;
	case 'M':
	    mkdev_flag++;
	    get_flag = 0;
	    break;
	case 'V':
	    fprintf(stderr, "setrocket version %s\n", VERSION_STR);
	    exit(0);
	default:
	    usage();
	}
    }
    if (get_flag) {
	argv += optind;
	while (*argv)
	    get_rocket(*argv++);
	exit(0);
    }
    if (argc == optind)
	usage();
    if (mkdev_flag) {
	mkdev_rocket(argv[optind]);
	exit(0);
    }
    if (argc-optind == 1)
	get_rocket(argv[optind]);
    else
	set_rocket(argv[optind], argv+optind+1);
    exit(0);
}

