/******************************************************************
 *
 *  unity.c
 *
 *  Unity accessor functions.
 *
 *******************************************************************/

#include <linux/kernel.h>
#include <linux/module.h> 
#include <linux/fs.h> 
#include <linux/pci.h>
#include <linux/termios.h>
#include <asm/uaccess.h>
#include <linux/sched.h>
#include "unity.h"

/*  Use of a global here may be non-recursive */
static unsigned short sval;
static unsigned int lval;

/*  Macros to set/clear a bit in a PCI memory word (short - 16 bits) */
#define SETBITW(addr, bit)  \
do { \
  sval = READW(addr);  \
  WRITEW(sval | bit, addr); \
} while (0)

#define CLEARBITW(addr, bit)  \
do { \
  sval = READW(addr);  \
  WRITEW(sval & ~bit, addr); \
} while (0)

/*  Macros to set/clear a bit in a PCI memory double word (long - 32 bits) */
#define SETBITL(addr, bit)  \
do { \
  lval = READL(addr);  \
  WRITEL(lval | bit, addr); \
} while (0)

#define CLEARBITL(addr, bit)  \
do { \
  lval = READL(addr);  \
  WRITEL(lval & ~bit, addr); \
} while (0)


#if (0)  // debug version kdr, feel free to muck around here...

/*  Channel microcode */
#define UCODESIZE 0x3F

static unsigned char uCode[UCODESIZE] = {
	0xF6, 0x8B, 0xE2, 0x86,     /*  0x00 */
	0x30, 0x40, 0x66, 0x19,     /*  0x04 */
	0x0A, 0x31, 0x40, 0x67,     /*  0x08 */
	0x19, 0x0a, 0x86, 0x01,     /*  0x0c */
	0x40, 0x11, 0x19, 0x0A,     /*  0x10 */
	0x00, 0x40, 0x13, 0x19,     /*  0x14 */
	0x0A, 0xFA, 0x83, 0x01,     /*  0x18 */
	0x0A, 0x00, 0x0A, 0x41,     /*  0x1C */
	0xFF, 0x89, 0xC2, 0x66,     /*  0x20 */
	0x86, 0x81, 0x91, 0x40,     /*  0x24 */
	0x65, 0x8E, 0x89, 0xC2,     /*  0x28 */
	0x66, 0x86, 0x81, 0x88,     /*  0x2C */
	0x40, 0x65, 0x85, 0x84,     /*  0x30 */
	0x00, 0x82, 0x0A, 0x08,     /*  0x34 */
	0x0A, 0x0A, 0x0A, 0x0A,     /*  0x38 */
	0x0A, 0x08, 0x0A            /*  0x3C */
};


#else   //  The release version


/*  Channel microcode */
#define UCODESIZE 0x3F

static unsigned char uCode[UCODESIZE] = {
	0xF6, 0x8B, 0x9D, 0xC4,     /*  0x00 */
	0x13, 0xC4, 0x11, 0x98,     /*  0x04 */
	0x20, 0x0A, 0x21, 0x0A,     /*  0x08 */
	0x8D, 0xFA, 0x86, 0x01,     /*  0x0c */
	0x40, 0x11, 0x19, 0x0A,     /*  0x10 */
	0x00, 0x40, 0x13, 0x19,     /*  0x14 */
	0x0A, 0xFA, 0x83, 0x01,     /*  0x18 */
	0x0A, 0x00, 0x0A, 0x41,     /*  0x1C */
	0xFF, 0x89, 0xC2, 0x66,     /*  0x20 */
	0x86, 0x81, 0x91, 0x40,     /*  0x24 */
	0x65, 0x8E, 0x89, 0xC2,     /*  0x28 */
	0x66, 0x86, 0x81, 0x88,     /*  0x2C */
	0x40, 0x65, 0x85, 0x84,     /*  0x30 */
	0x00, 0x82, 0x0A, 0x08,     /*  0x34 */
	0x0A, 0x0A, 0x0A, 0x0A,     /*  0x38 */
	0x0A, 0x08, 0x0A            /*  0x3C */
};

#endif

/*  
 *  Sets the clock prescale value.
 *  Inputs:   CtlP pointer to controller struct, prescale value. 
 */
void uSetClockPrescale(CONTROLLER_T *CtlP, unsigned int prescale)
{
    WRITEL(prescale & 0xFF, CtlP->pBasePciMemory + ASIC_CLOCK_PRESCALER);
    if (CtlP->has_2nd_asic)
	WRITEL(prescale & 0xFF, CtlP->pBasePciMemory + 0x1000 + ASIC_CLOCK_PRESCALER);
}

/*  
 *  Sets the TDM clock rate value.
 *  Inputs:   CtlP pointer to controller struct. 
 */
void uSetTDMClockRate(CONTROLLER_T *CtlP)
{
	SETBITW(CtlP->pBasePciMemory + ASIC_CONFIGURATION, TDM_CLK_SPEED1);
	if(CtlP->has_2nd_asic)
        	SETBITW(CtlP->pBasePciMemory + 0x1000 + ASIC_CONFIGURATION, TDM_CLK_SPEED1);
	CLEARBITW(CtlP->pBasePciMemory + ASIC_CONFIGURATION, TDM_CLK_SPEED0);
	if(CtlP->has_2nd_asic)
		CLEARBITW(CtlP->pBasePciMemory + 0x1000 + ASIC_CONFIGURATION, TDM_CLK_SPEED0);
}

/*  
 *   Performs a global reset.  Do not need to clear the bit after setting it, or 
 *   wait after.  (BKA: not true!  we definitely need a delay, clear...maybe)
 *   Inputs:   CtlP pointer to controller struct.
 */
void uGlobalReset(CONTROLLER_T *CtlP)
{
    WRITEW(1, CtlP->pBasePciMemory + GLOBAL_COMMAND);
    if (CtlP->has_2nd_asic) 
	WRITEW(1, CtlP->pBasePciMemory + 0x1000 + GLOBAL_COMMAND);

    //new
    set_current_state(TASK_INTERRUPTIBLE);//BKA
    schedule_timeout(HZ / 10);//BKA
    /*the following also added back in for same reason
    WRITEW(0, CtlP->pBasePciMemory + GLOBAL_COMMAND);
    if (CtlP->has_2nd_asic) 
	WRITEW(0, CtlP->pBasePciMemory + 0x1000 + GLOBAL_COMMAND);
    */
    return;
}

/*  
 *  Resets the channel.  
 *  Inputs:   ChP pointer to channel struct, baud divisor value. 
 */
void uChannelReset(CHANNEL_T *ChP)
{
    SETBITL(ChP->pBaseRegister + UART_CTRL, RESET_CHANNEL);
 }


/*  
 *  Sets the baud divisor on a channel.
 *  Inputs:   ChP pointer to channel struct, baud divisor value. 
 */
void uSetBaudDivisor(CHANNEL_T *ChP, unsigned int divisor)
{
    WRITEW(divisor, ChP->pBaseRegister + BAUD_RATE_GENERATOR);
}

/*  
 *  Sets the Intercharacter timeout reload value.
 *  Inputs:   ChP pointer to channel struct, timeout value in ms. 
 */
void uSetIntercharValue(CHANNEL_T *ChP, unsigned int val)
{
    WRITEL(val, ChP->pBaseRegister + INTERCHAR_RELOAD_VAL);
}

/*  
 *  Flushes the TX/RX FIFOs.
 *  Inputs:   ChP pointer to channel struct. 
 */
void uFlushTxFifo(CHANNEL_T *ChP)
{
    SETBITL(ChP->pBaseRegister + UART_CTRL, FLUSH_XMIT_FIFO);
    CLEARBITL(ChP->pBaseRegister + UART_CTRL, FLUSH_XMIT_FIFO);
}

void uFlushRxFifo(CHANNEL_T *ChP)
{
    SETBITL(ChP->pBaseRegister + UART_CTRL, FLUSH_RCV_FIFO);
    CLEARBITL(ChP->pBaseRegister + UART_CTRL, FLUSH_RCV_FIFO);
}

void uFlushPriorityFifo(CHANNEL_T *ChP)
{
    SETBITL(ChP->pBaseRegister + UART_CTRL, FLUSH_PRIORITY_FIFO);
    CLEARBITL(ChP->pBaseRegister + UART_CTRL, FLUSH_PRIORITY_FIFO);
}

/*  
 *  Send/Stop sending a transmit BREAK signal.
 *  Inputs:   ChP pointer to channel struct. 
 */
void uSendBreak(CHANNEL_T *ChP)
{
    SETBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, BREAK_ON);
}

void uClrBreak(CHANNEL_T *ChP)
{
    CLEARBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, BREAK_ON);
}


/*  
 *  Set/clear the DTR output pin.
 *  Inputs:   ChP pointer to channel struct. 
 */
void uSetDtr(CHANNEL_T *ChP)
{
    SETBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, SET_DTR_ACTIVE);
}

void uClrDtr(CHANNEL_T *ChP)
{
    CLEARBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, SET_DTR_ACTIVE);
}


/*  
 *  Set/clear RTS output pin.
 *  Inputs:   ChP pointer to channel struct. 
 */
void uSetRts(CHANNEL_T *ChP)
{
    SETBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, SET_RTS_ACTIVE);
}

void uClrRts(CHANNEL_T *ChP)
{
    CLEARBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, SET_RTS_ACTIVE);
}

/*  
 *  Enable/disable CTS flow control.
 *   Inputs:   ChP pointer to channel struct. 
 */
void uEnCtsFlowCtrl(CHANNEL_T *ChP)
{
    SETBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, CTS_FLOW_EN);
}

void uDisCtsFlowCtrl(CHANNEL_T *ChP)
{
    CLEARBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, CTS_FLOW_EN);
}


/*  
 *  Enable/disable RTS output flow control.
 *   Inputs:   ChP pointer to channel struct. 
 */
void uEnRtsFlowCtrl(CHANNEL_T *ChP)
{
    SETBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, RTS_FLOW_EN);
}

void uDisRtsFlowCtrl(CHANNEL_T *ChP)
{
    CLEARBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, RTS_FLOW_EN);
}


/*  
 *  Enable/disable parity.
 *  Inputs:   ChP pointer to channel struct. 
 */
void uEnParity(CHANNEL_T *ChP)
{
    SETBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, PARITY_ON);
}

void uDisParity(CHANNEL_T *ChP)
{
    CLEARBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, PARITY_ON);
}

/*  
 *  Enable/disable forced parity.
 *  Inputs:   ChP pointer to channel struct. 
 */
void uEnForcedParity(CHANNEL_T *ChP)
{
    SETBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, FORCE_PARITY);
}

void uDisForcedParity(CHANNEL_T *ChP)
{
    CLEARBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, FORCE_PARITY);
}


/*  
 *  Enable/disable RTS toggle.
 *  Inputs:   ChP pointer to channel struct. 
 */
void uEnRTSToggle(CHANNEL_T *ChP)
{
    SETBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, RTS_TOGGLE);
}

void uDisRTSToggle(CHANNEL_T *ChP)
{
    CLEARBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, RTS_TOGGLE);
}

/*  
 *  Start/stop the Receive Processor.  Procedure is to have the processor stopped, 
 *  load ucode, then start the processor.  To subsequently disable the recieve 
 *  processor, don't use this bit.  Instead use the uDisRxFIFO()/uEnRxFIFO() functions
 *  which uses the microcode flag.
 *
 *  Background:  Clearing this flag causes the received bytes to be handled by
 *  a 16550 UART core rather than the receive processor.  This only will work for
 *  channel 0, and should never be used by the drivers.  It is for ASIC debug only.
 *  Setting this bit causes received serial bytes to be handled by the receive processor.
 *
 *  Inputs:   ChP pointer to channel struct. 
 */
void uStartRxProcessor(CHANNEL_T *ChP)
{
    SETBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, RCV_PROC_EN);
}

void uStopRxProcessor(CHANNEL_T *ChP)
{
    CLEARBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, RCV_PROC_EN);
}

/*  
 *  Enable/disable the RX FIFO.
 *  Inputs:   ChP pointer to channel struct. 
 */
void uEnRxFIFO(CHANNEL_T *ChP)
{                                               
    WRITEB(URP_RCV_FIFO_ENABLE, ChP->ucodeAddr + URP_RCV_FIFO);
}

void uDisRxFIFO(CHANNEL_T *ChP)
{
    WRITEB(URP_RCV_FIFO_DISABLE, ChP->ucodeAddr + URP_RCV_FIFO);
}

/*  
 *  Enable/disable the transmitter.
 *  Inputs:   ChP pointer to channel struct. 
 */
void uEnTransmit(CHANNEL_T *ChP)
{
    SETBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, XMIT_EN);
}

void uDisTransmit(CHANNEL_T *ChP)
{
    CLEARBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, XMIT_EN);
}


/*  
 *  Enable/disable the strip nulls function in the receive processor.
 *  Inputs:   ChP pointer to channel struct. 
 */
void uEnStripNulls(CHANNEL_T *ChP)
{
    WRITEB(URP_STRIP_NULLS_ENABLE, ChP->ucodeAddr + URP_STRIP_NULLS);
}

void uDisStripNulls(CHANNEL_T *ChP)
{
    WRITEB(URP_STRIP_NULLS_DISABLE, ChP->ucodeAddr + URP_STRIP_NULLS);
}


/*  
 *  Enable/disable TX software flow control.
 *  Inputs:   ChP pointer to channel struct. 
 */
void uEnSoftFlowCtrl(CHANNEL_T *ChP)
{
    WRITEB(URP_TX_SW_FLOW_CONTROL_ENABLE, ChP->ucodeAddr + URP_TX_SW_FLOW_CONTROL);
    WRITEB(URP_RX_SW_FLOW_CONTROL_ENABLE, ChP->ucodeAddr + URP_RX_SW_FLOW_CONTROL);
}

void uDisSoftFlowCtrl(CHANNEL_T *ChP)
{
    WRITEB(URP_TX_SW_FLOW_CONTROL_DISABLE, ChP->ucodeAddr + URP_TX_SW_FLOW_CONTROL);
    WRITEB(URP_RX_SW_FLOW_CONTROL_DISABLE, ChP->ucodeAddr + URP_RX_SW_FLOW_CONTROL);
}

/*  
 *  Sets the mask in the receive processor.
 *  Inputs:   ChP pointer to channel struct, mask. 
 */
void uSetMask(CHANNEL_T *ChP, unsigned char mask)
{
    WRITEB(mask, ChP->ucodeAddr + URP_MASK);
}

// kdr  Perhaps this should be moved into the application...

/*  Defines used in the Receive Processor Compare function */
#define COMPARE_SAVE_BYTE    0
#define COMPARE_REPLACE_BYTE 1
#define COMPARE_DISCARD_BYTE 2

/*  
 *  Enable/disable compare #1 & 2.
 *  Inputs:   ChP pointer to channel struct. 
 */
void uEnCompare1(CHANNEL_T *ChP, unsigned char compareAction, unsigned char compareChar, unsigned char replaceChar)
{
    switch (compareAction) {
    case COMPARE_DISCARD_BYTE:
	WRITEB(compareChar, ChP->ucodeAddr + URP_COMPARE_BYTE1);
	WRITEB(URP_COMPARE_ACTION1_DISCARD, ChP->ucodeAddr + URP_COMPARE_ACTION1);
	WRITEB(URP_COMPARE1_ENABLE, ChP->ucodeAddr + URP_COMPARE1);
	WRITEB(URP_COMPARE_IRQ1_ENABLE, ChP->ucodeAddr + URP_COMPARE_IRQ1);
	break;
    case COMPARE_REPLACE_BYTE:
	WRITEB(compareChar, ChP->ucodeAddr + URP_COMPARE_BYTE1);
	WRITEB(replaceChar, ChP->ucodeAddr + URP_REPLACE_BYTE1);
	WRITEB(URP_COMPARE_ACTION1_REPLACE, ChP->ucodeAddr + URP_COMPARE_ACTION1);
	WRITEB(URP_COMPARE1_ENABLE, ChP->ucodeAddr + URP_COMPARE1);
	WRITEB(URP_COMPARE_IRQ1_ENABLE, ChP->ucodeAddr + URP_COMPARE_IRQ1);
	break;
    case COMPARE_SAVE_BYTE:
    default: 
	WRITEB(compareChar, ChP->ucodeAddr + URP_COMPARE_BYTE1);
	WRITEB(URP_COMPARE_ACTION1_SAVE, ChP->ucodeAddr + URP_COMPARE_ACTION1);
	WRITEB(URP_COMPARE1_ENABLE, ChP->ucodeAddr + URP_COMPARE1);
	WRITEB(URP_COMPARE_IRQ1_ENABLE, ChP->ucodeAddr + URP_COMPARE_IRQ1);
	break;
    }
}

void uEnCompare2(CHANNEL_T *ChP, unsigned char compareAction, unsigned char compareChar, unsigned char replaceChar)
{
    switch (compareAction) {
    case COMPARE_DISCARD_BYTE:
	WRITEB(compareChar, ChP->ucodeAddr + URP_COMPARE_BYTE2);
	WRITEB(URP_COMPARE_ACTION2_DISCARD, ChP->ucodeAddr + URP_COMPARE_ACTION2);
	WRITEB(URP_COMPARE2_ENABLE, ChP->ucodeAddr + URP_COMPARE2);
	WRITEB(URP_COMPARE_IRQ2_ENABLE, ChP->ucodeAddr + URP_COMPARE_IRQ2);
	break;
    case COMPARE_REPLACE_BYTE:
	WRITEB(compareChar, ChP->ucodeAddr + URP_COMPARE_BYTE2);
	WRITEB(replaceChar, ChP->ucodeAddr + URP_REPLACE_BYTE2);
	WRITEB(URP_COMPARE_ACTION2_REPLACE, ChP->ucodeAddr + URP_COMPARE_ACTION2);
	WRITEB(URP_COMPARE2_ENABLE, ChP->ucodeAddr + URP_COMPARE2);
	WRITEB(URP_COMPARE_IRQ2_ENABLE, ChP->ucodeAddr + URP_COMPARE_IRQ2);
	break;
    case COMPARE_SAVE_BYTE:
    default: 
	WRITEB(compareChar, ChP->ucodeAddr + URP_COMPARE_BYTE2);
	WRITEB(URP_COMPARE_ACTION2_SAVE, ChP->ucodeAddr + URP_COMPARE_ACTION2);
	WRITEB(URP_COMPARE2_ENABLE, ChP->ucodeAddr + URP_COMPARE2);
	WRITEB(URP_COMPARE_IRQ2_ENABLE, ChP->ucodeAddr + URP_COMPARE_IRQ2);
	break;
    }
}

void uDisCompare1(CHANNEL_T *ChP)
{
    WRITEB(URP_COMPARE1_DISABLE, ChP->ucodeAddr + URP_COMPARE1);
    WRITEB(URP_COMPARE_IRQ1_DISABLE, ChP->ucodeAddr + URP_COMPARE_IRQ1);
}

void uDisCompare2(CHANNEL_T *ChP)
{
    WRITEB(URP_COMPARE2_DISABLE, ChP->ucodeAddr + URP_COMPARE2);
    WRITEB(URP_COMPARE_IRQ2_DISABLE, ChP->ucodeAddr + URP_COMPARE_IRQ2);
}

/*  
 *   Set the Rx/Tx XOFF/XON flow control characters
 *  Inputs:   ChP pointer to channel struct, the character. 
 */
void uSetTxXOFFChar(CHANNEL_T *ChP, char ch)
{
    WRITEB(ch, ChP->ucodeAddr + URP_XMIT_XOFF_CHAR);
}

void uSetTxXONChar(CHANNEL_T *ChP, char ch)
{
    WRITEB(ch, ChP->ucodeAddr + URP_XMIT_XON_CHAR);
}

void uSetRxXOFFChar(CHANNEL_T *ChP, char ch)
{
    WRITEB(ch, ChP->ucodeAddr + URP_RCV_XOFF_CHAR);
}

void uSetRxXONChar(CHANNEL_T *ChP, char ch)
{
    WRITEB(ch, ChP->ucodeAddr + URP_RCV_XON_CHAR);
}


/*  
 *  Set stop bits to 1 or 2
 *  Inputs:   ChP pointer to channel struct. 
 */
void uSetStop1(CHANNEL_T *ChP)
{
    CLEARBITL(ChP->pBaseRegister + UART_CTRL, STOP_BITS);
}

void uSetStop2(CHANNEL_T *ChP)
{
    SETBITL(ChP->pBaseRegister + UART_CTRL, STOP_BITS);
}


/*  
 *  Set TX FIFO trigger level
 *  Inputs:   ChP pointer to channel struct, level sets the number of characters in 
 *            FIFO at which the  interrupt will be generated.  
 *            -  UXMIT_FIFO_TRIG_DISABLED  disabled the XMIT FIFO Trigger
 *            -  UXMIT_FIFO_TRIG_1_8       1/8 full
 *            -  UXMIT_FIFO_TRIG_1_4       1/4 full
 *            -  UXMIT_FIFO_TRIG_1_2       1/2 full
 */
void uSetTxTrigger(CHANNEL_T *ChP, unsigned int level)
{
    unsigned int aval;

    aval = READL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL);  
    aval &= ~UXMIT_FIFO_TRIG_MASK;
    WRITEL(aval | level, ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL); 
}

/*  
 *  Set RX FIFO trigger level
 *  Inputs:   ChP pointer to channel struct, level sets the number of characters in 
 *            FIFO at which the  interrupt will be generated.  
 *            -  URCV_FIFO_TRIG_DISABLED  disabled the RCV FIFO Trigger
 *            -  URCV_FIFO_TRIG_1_CHAR    1 Character
 *            -  URCV_FIFO_TRIG_1_2       1/2 full
 *            -  URCV_FIFO_TRIG_7_8       7/8 full
 */
void uSetRxTrigger(CHANNEL_T *ChP, unsigned int level)
{
    unsigned int aval;

    aval = READL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL);  
    aval &= ~URCV_FIFO_TRIG_MASK;
    WRITEL(aval | level, ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL); 
}
   

/*  
 *  Set even/odd parity
 *  Inputs:   ChP pointer to channel struct. 
 */
void uSetEvenParity(CHANNEL_T *ChP)
{
    SETBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, EVEN_PARITY);
}

void uSetOddParity(CHANNEL_T *ChP)
{
    CLEARBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, EVEN_PARITY);
}


/*  
 *  Functions to set number of data bits
 *  Inputs:   ChP pointer to channel struct. 
 */
void uSetDataBits(CHANNEL_T *ChP, unsigned int level)
{
    unsigned int aval;

    aval = READL(ChP->pBaseRegister + UART_CTRL);  
    aval &= ~DATABITS_MASK;
    WRITEL(aval | level, ChP->pBaseRegister + UART_CTRL); 
}

void uSetData5(CHANNEL_T *ChP)
{
    uSetDataBits(ChP, DATABITS_5);
}

void uSetData6(CHANNEL_T *ChP)
{
    uSetDataBits(ChP, DATABITS_6);
}

void uSetData7(CHANNEL_T *ChP)
{
    uSetDataBits(ChP, DATABITS_7);
}

void uSetData8(CHANNEL_T *ChP)
{
    uSetDataBits(ChP, DATABITS_8);
}


/*  
 *  Get the number of data bytes in the TX/RX FIFOs
 *  Inputs:   ChP pointer to channel struct. 
 *  Returns:  FIFO count
 */
unsigned short uGetTxCnt(CHANNEL_T *ChP)
{
    return READW(ChP->pBaseRegister + XMIT_FIFO_COUNT);
}

unsigned short uGetRxCnt(CHANNEL_T *ChP)
{
    return READW(ChP->pBaseRegister + RCV_FIFO_COUNT);
}

/*  
 *  Set the RS interface mode.
 *  Inputs:   ChP pointer to channel struct, mode byte indicating the mode.
 *  NOTE:  Different modes than AIOPIC, see URSxxx in unity.h.
 */
void uSetInterfaceMode(CHANNEL_T *ChP, unsigned int mode)
{
    unsigned int aval;

    aval = READL(ChP->pBaseRegister + UART_CTRL);  
    aval &= ~URS_MASK;
    WRITEL(aval | mode, ChP->pBaseRegister + UART_CTRL); 
}

/*
 *  Enable/disable tranmitter loopback mode.
 *  Inputs:   ChP pointer to channel struct
 */
void uEnLoopbackMode(CHANNEL_T *ChP)
{
    SETBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, LOCAL_LOOPBACK);
}

void uDisLoopbackMode(CHANNEL_T *ChP)
{
    CLEARBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, LOCAL_LOOPBACK);
}


/*
 *  Set RTS toggle polarity to active low/high.
 *  Inputs:   ChP pointer to channel struct
 */
void uRtsTogglePolarityHigh(CHANNEL_T *ChP)
{
    SETBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, RTS_TOGGLE_POLARITY);
}

void uRtsTogglePolarityLow(CHANNEL_T *ChP)
{
    CLEARBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, RTS_TOGGLE_POLARITY);
}


/*
 *  Enable/disable tranmitter DSR flow control.
 *  Inputs:   ChP pointer to channel struct
 */
void uEnDsrFlowCtrl(CHANNEL_T *ChP)
{
    SETBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, DSR_FLOW_EN);
}

void uDisDsrFlowCtrl(CHANNEL_T *ChP)
{
    CLEARBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, DSR_FLOW_EN);
}


/*
 *  Enable/disable receiver DTR flow control.
 *  Inputs:   ChP pointer to channel struct
 */
void uEnDtrFlowCtrl(CHANNEL_T *ChP)
{
    SETBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, DTR_FLOW_EN);
}

void uDisDtrFlowCtrl(CHANNEL_T *ChP)
{
    CLEARBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, DTR_FLOW_EN);
}


/*
 *  Returns the RI line state.
 *  Returns:  1 if RI is set, else 0.
 */
int uGetChanRI(CHANNEL_T * ChP)
{
    return (READL(ChP->pBaseRegister + INT_ID_CHAN_STATUS) & CHAN_STATUS_RI_ACTIVE);
}


/*
 *  Transmits a serial byte/word/long.  No check of FIFO is done first, that must be done
 *  in the calling function.
 *  Inputs:   ChP pointer to channel struct, Data to send
 */
void uXmitByte(CHANNEL_T *ChP, unsigned char pval)
{
    WRITEB(pval, ChP->pBaseRegister + XMIT_DATA_BYTE);
}

void uXmitWord(CHANNEL_T *ChP, unsigned short pval)
{
    WRITEW(pval, ChP->pBaseRegister + XMIT_DATA_WORD);
}

void uXmitDword(CHANNEL_T *ChP, unsigned int pval)
{
    WRITEL(pval, ChP->pBaseRegister + XMIT_DATA_BLOCK);
}

/*
 *  Transmits a serial byte/word/long using the priority buffer.
 *  Inputs:   ChP pointer to channel struct, Data to send
 */
void uXmitBytePriority(CHANNEL_T *ChP, unsigned char pval)
{
    WRITEB(pval, ChP->pBaseRegister + PRIORITY_XMIT_BYTE);
}

void uXmitWordPriority(CHANNEL_T *ChP, unsigned short pval)
{
    WRITEW(pval, ChP->pBaseRegister + PRIORITY_XMIT_WORD);
}

void uXmitDwordPriority(CHANNEL_T *ChP, unsigned int pval)
{
    WRITEW((unsigned short)(pval & 0xFFFF), ChP->pBaseRegister + PRIORITY_XMIT_WORD);
    WRITEW((unsigned short)((pval & 0xFFFF0000) >> 16), ChP->pBaseRegister + PRIORITY_XMIT_WORD);
}

/*
 *  Reads a serial byte/word/long.  No check of FIFO is done first, that must be done
 *  in the calling function.
 *  Inputs:   ChP pointer to channel struct, Data to send
 */
unsigned short uRxByteStatus(CHANNEL_T *ChP)
{
    return READW(ChP->pBaseRegister + RCV_DATA_BYTE_STATUS);
}

unsigned short uRxWord(CHANNEL_T *ChP)
{
    return READW(ChP->pBaseRegister + RCV_DATA_WORD);
}

unsigned int uRxDword(CHANNEL_T *ChP)
{
    return READL(ChP->pBaseRegister + RCV_DATA_BLOCK);
}

/*
 *  Enable/Disable global interrupts.
 *  Inputs:   ChP pointer to channel struct
 */
void uEnableGlobalInterrupts(CONTROLLER_T *CtlP)
{
    unsigned int intMask = (PCI_INTERRUPT_ENABLE | PCI_WISHBONE_ERROR_ENABLE | PCI_ERROR_ENABLE | PCI_PARITY_ERROR_ENABLE | PCI_SYSTEM_ERROR_ENABLE);

    SETBITL(CtlP->pBasePciRegion0 + PCI_INTERRUPT_CONTROL, intMask);
}

void uDisableGlobalInterrupts(CONTROLLER_T *CtlP)
{
    WRITEL(0, CtlP->pBasePciRegion0 + PCI_INTERRUPT_CONTROL);
}

int uGlobalInterruptsAreEnabled(CONTROLLER_T *CtlP)
{
    return READL(CtlP->pBasePciRegion0 + PCI_INTERRUPT_CONTROL) & PCI_INTERRUPT_ENABLE;
}

/*
 *  Enable/Disable AIOP interrupts.
 *  Inputs:   ChP pointer to channel struct
 */
void uEnableAiopInterrupts(CONTROLLER_T *CtlP)
{
    SETBITL(CtlP->pBasePciMemory + ASIC_INT_STATUS, AIOP_INT_EN);
}

void uDisableAiopInterrupts(CONTROLLER_T *CtlP)
{
    CLEARBITL(CtlP->pBasePciMemory + ASIC_INT_STATUS, AIOP_INT_EN);
}

/*  
 *  Clears global hardware interrupts.  This is done in Region 0, the Flextronics core. 
 *  Inputs:   CtrP pointer to controller struct
 */
void uClearGlobalInts(CONTROLLER_T *CtlP)
{
    WRITEL(0, CtlP->pBasePciRegion0 + PCI_INTERRUPT_STATUS);
}

/*  
 *  Enable/disable the AIOP interrupts
 *  Inputs:   ChP pointer to channel struct
 */
void uEnXmitFifoEmptyInt(CHANNEL_T *ChP)
{
    SETBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, XMIT_FIFO_INT_EN);
}

void uDisXmitFifoEmptyInt(CHANNEL_T *ChP)
{
    CLEARBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, XMIT_FIFO_INT_EN);
}

void uEnXmitShiftRegEmptyInt(CHANNEL_T *ChP)
{
    SETBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, XMIT_SHIFTREG_INT_EN);
}

void uDisXmitShiftRegEmptyInt(CHANNEL_T *ChP)
{
    CLEARBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, XMIT_SHIFTREG_INT_EN);
}

void uEnInterCharTimeoutInt(CHANNEL_T *ChP)
{
    SETBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, INTERCHAR_TIMEOUT_INT_EN);
}

void uDisInterCharTimeoutInt(CHANNEL_T *ChP)
{
    CLEARBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, INTERCHAR_TIMEOUT_INT_EN);
}

void uEnInterCharTimeoutStatus(CHANNEL_T *ChP)
{
    SETBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, INTERCHAR_STATUS_EN);
}

void uDisInterCharTimeoutStatus(CHANNEL_T *ChP)
{
    CLEARBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, INTERCHAR_STATUS_EN);
}

void uEnAddRcvdInt(CHANNEL_T *ChP)
{
    SETBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, ADDRESS_INT_EN);
}

void uDisAddRcvdInt(CHANNEL_T *ChP)
{
    CLEARBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, ADDRESS_INT_EN);
}

void uEnRcvDataReadyTimeoutInt(CHANNEL_T *ChP)
{
    SETBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, RCV_DATA_RDY_INT_EN);
}

void uDisRcvDataReadyTimeoutInt(CHANNEL_T *ChP)
{
    CLEARBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, RCV_DATA_RDY_INT_EN);
}

void uEnSpecialReceiveInt(CHANNEL_T *ChP)
{
    SETBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, SPECIAL_RCV_INT_EN);
}

void uDisSpecialReceiveInt(CHANNEL_T *ChP)
{
    CLEARBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, SPECIAL_RCV_INT_EN);
}

void uEnModemChangeInt(CHANNEL_T *ChP)
{
    SETBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, MDM_CHANGE_INT_EN);
}

void uDisModemChangeInt(CHANNEL_T *ChP)
{
    CLEARBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, MDM_CHANGE_INT_EN);
}


/*  
 *  Reports whether this PCI card was the source of the interrupt.  Reads the Flextronics
 *  PCI Interrupt Status register, which also detects other PCI interrupts.
 *  Returns:  1 if yes, else 0
 */
int uInterruptActive(CONTROLLER_T *CtlP)
{
    unsigned int intStatus, intMask = (PCI_INTERRUPT_STATUS_BIT | PCI_WISHBONE_ERROR | PCI_ERROR | PCI_PARITY_ERROR | PCI_SYSTEM_ERROR);

    intStatus = READL(CtlP->pBasePciRegion0 + PCI_INTERRUPT_STATUS);
    if (intStatus & intMask)
	return 1;
    else
	return 0;
}

/*  
 *  Functions that report whether a particular interrupt has occurred
 *  Returns:  1 if yes, else 0
 */
int uDmaInterruptActive(CONTROLLER_T *CtlP)
{
    return READL(CtlP->pBasePciMemory + ASIC_INT_STATUS) & DMA_INT_ACTIVE;
}

int uAiop0InterruptActive(CONTROLLER_T *CtlP)
{
    return READL(CtlP->pBasePciMemory + ASIC_INT_STATUS) & AIOP0_INT_ACTIVE;
}

int uAiop1InterruptActive(CONTROLLER_T *CtlP)
{
    return READL(CtlP->pBasePciMemory + ASIC_INT_STATUS) & AIOP1_INT_ACTIVE;
}

int uAiop2InterruptActive(CONTROLLER_T *CtlP)
{
    return READL(CtlP->pBasePciMemory + ASIC_INT_STATUS) & AIOP2_INT_ACTIVE;
}

int uAiop3InterruptActive(CONTROLLER_T *CtlP)
{
    return READL(CtlP->pBasePciMemory + ASIC_INT_STATUS) & AIOP3_INT_ACTIVE;
}

int uAnyAiopInterrruptActive(CONTROLLER_T *CtlP)
{
    return READL(CtlP->pBasePciMemory + ASIC_INT_STATUS) & (AIOP0_INT_ACTIVE | AIOP1_INT_ACTIVE | AIOP2_INT_ACTIVE | AIOP3_INT_ACTIVE) ;
}

/*  
 *  Interrupt identification functions. Bits must be cleared if 
 *  set by writing a '1' to them.  The special receive is not immediately 
 *  cleared here, since it has multiple sources.
 *  Returns:  1 if yes, else 0
 */
int uInterCharTimeoutInt(CHANNEL_T *ChP)
{
    int ret;

    if ((ret = READL(ChP->pBaseRegister + INT_ID_CHAN_STATUS) & INT_INTERCHAR_TIMEOUT))
	WRITEL(INT_INTERCHAR_TIMEOUT, ChP->pBaseRegister + INT_ID_CHAN_STATUS);
    return ret;
}

int uXmitFifoTriggerInt(CHANNEL_T *ChP)
{
    int ret;

    if ((ret = READL(ChP->pBaseRegister + INT_ID_CHAN_STATUS) & INT_XMIT_FIFO_TRIGGER))
	WRITEL(INT_XMIT_FIFO_TRIGGER, ChP->pBaseRegister + INT_ID_CHAN_STATUS);
    return ret;
}

int uXmitShiftRegEmptyInt(CHANNEL_T *ChP)
{
    int ret;

    if ((ret = READL(ChP->pBaseRegister + INT_ID_CHAN_STATUS) & INT_XMIT_SHIFTREG_EMPTY))
	WRITEL(INT_XMIT_SHIFTREG_EMPTY, ChP->pBaseRegister + INT_ID_CHAN_STATUS);
    return ret;
}

int uAddrRcvdInt(CHANNEL_T *ChP)
{
    int ret;

    if ((ret = READL(ChP->pBaseRegister + INT_ID_CHAN_STATUS) & INT_ADDR_RCVD))
	WRITEL(INT_ADDR_RCVD, ChP->pBaseRegister + INT_ID_CHAN_STATUS);
    return ret;
}

int uRcvDataRdyInt(CHANNEL_T *ChP)
{
    int ret;

    if ((ret = READL(ChP->pBaseRegister + INT_ID_CHAN_STATUS) & INT_RCV_DATA_RDY))
	WRITEL(INT_RCV_DATA_RDY, ChP->pBaseRegister + INT_ID_CHAN_STATUS);
    return ret;
}

int uIntRiChange(CHANNEL_T *ChP)
{
    int ret;

    if ((ret = READL(ChP->pBaseRegister + INT_ID_CHAN_STATUS) & INT_RI_CHANGE))
	WRITEL(INT_RI_CHANGE, ChP->pBaseRegister + INT_ID_CHAN_STATUS);
    return ret;
}

int uRcvFifoTriggerInt(CHANNEL_T *ChP)
{
    int ret;

    if ((ret = READL(ChP->pBaseRegister + INT_ID_CHAN_STATUS) & INT_RCV_FIFO_TRIGGER))
	WRITEL(INT_RCV_FIFO_TRIGGER, ChP->pBaseRegister + INT_ID_CHAN_STATUS);
    return ret;
}

int uXmitFifoEmptyInt(CHANNEL_T *ChP)
{
    int ret;

    if ((ret = READL(ChP->pBaseRegister + INT_ID_CHAN_STATUS) & INT_XMIT_FIFO_EMPTY))
	WRITEL(INT_XMIT_FIFO_EMPTY, ChP->pBaseRegister + INT_ID_CHAN_STATUS);
    return ret;
}

/* 
 *  Retrieve/clear the Special Receive interrupt flag.  Should read it, them 
 *  determine and clear the source flags (parity, framing, etc), then clear this bit.
 */
int uSpecialReceiveInt(CHANNEL_T *ChP)
{
    return (READL(ChP->pBaseRegister + INT_ID_CHAN_STATUS) & INT_SPECIAL_RECEIVE);
}

void uClrSpecialReceiveInt(CHANNEL_T *ChP)
{
    WRITEL(INT_SPECIAL_RECEIVE, ChP->pBaseRegister + INT_ID_CHAN_STATUS);
}

int uIntCdChange(CHANNEL_T *ChP)
{
    int ret;

    if ((ret = READL(ChP->pBaseRegister + INT_ID_CHAN_STATUS) & INT_CD_CHANGE))
	WRITEL(INT_CD_CHANGE, ChP->pBaseRegister + INT_ID_CHAN_STATUS);
    return ret;
}

int uIntCtsChange(CHANNEL_T *ChP)
{
    int ret = 0;

    if ((ret = READL(ChP->pBaseRegister + INT_ID_CHAN_STATUS) & INT_CTS_CHANGE))
	WRITEL(INT_CTS_CHANGE, ChP->pBaseRegister + INT_ID_CHAN_STATUS);
    return ret;
}

int uIntDsrChange(CHANNEL_T *ChP)
{
    int ret;

    if ((ret = READL(ChP->pBaseRegister + INT_ID_CHAN_STATUS) & INT_DSR_CHANGE))
	WRITEL(INT_DSR_CHANGE, ChP->pBaseRegister + INT_ID_CHAN_STATUS);

    return ret;
}


/*
 * Get/Clr channel status register 
 */
unsigned int uGetChanStatus(CHANNEL_T *ChP)
{
    return READL(ChP->pBaseRegister + INT_ID_CHAN_STATUS);
}

int uClrChanStatus(CHANNEL_T *ChP)
{
    WRITEL(CHAN_STATUS_RCV_OVERFLOW, ChP->pBaseRegister + INT_ID_CHAN_STATUS);
    WRITEL(CHAN_STATUS_BREAK_DETECTED, ChP->pBaseRegister + INT_ID_CHAN_STATUS);
    WRITEL(CHAN_STATUS_FRAMING_ERROR, ChP->pBaseRegister + INT_ID_CHAN_STATUS);
    WRITEL(CHAN_STATUS_PARITY_ERROR, ChP->pBaseRegister + INT_ID_CHAN_STATUS);
    return 0;
}


/*  
 *  Channel status functions (offset 0x10).  Bits 0-7 are read only.  Bits 8-15
 *  must be cleared if set by writing a '1' to them.  Exception - Priority Transmit
 *  FIFO Full bit is read only.
 *  Returns:  1 if yes, else 0
 */
int uPriorityBufferFull(CHANNEL_T *ChP)
{
    return READL(ChP->pBaseRegister + INT_ID_CHAN_STATUS) & CHAN_STATUS_PRIORITY_BUFFER_FULL;
}

int uChanStatusRcvOverflow(CHANNEL_T *ChP)
{
    int ret;

    if ((ret = READL(ChP->pBaseRegister + INT_ID_CHAN_STATUS) & CHAN_STATUS_RCV_OVERFLOW))
	WRITEL(CHAN_STATUS_RCV_OVERFLOW, ChP->pBaseRegister + INT_ID_CHAN_STATUS);
    return ret;
}

int uChanStatusCompare2(CHANNEL_T *ChP)
{
    int ret;

    if ((ret = READL(ChP->pBaseRegister + INT_ID_CHAN_STATUS) & CHAN_STATUS_COMPARE2)) 
	WRITEL(CHAN_STATUS_COMPARE2, ChP->pBaseRegister + INT_ID_CHAN_STATUS);
    return ret;
}

int uChanStatusCompare1(CHANNEL_T *ChP)
{
    int ret;

    if ((ret = READL(ChP->pBaseRegister + INT_ID_CHAN_STATUS) & CHAN_STATUS_COMPARE1))
	WRITEL(CHAN_STATUS_COMPARE1, ChP->pBaseRegister + INT_ID_CHAN_STATUS);
    return ret;
}

int uChanStatusBreakDetected(CHANNEL_T *ChP)
{
    int ret;

    if ((ret = READL(ChP->pBaseRegister + INT_ID_CHAN_STATUS) & CHAN_STATUS_BREAK_DETECTED))
	WRITEL(CHAN_STATUS_BREAK_DETECTED, ChP->pBaseRegister + INT_ID_CHAN_STATUS);
    return ret;
}

int uChanStatusFramingError(CHANNEL_T *ChP)
{
    int ret;

    if ((ret = READL(ChP->pBaseRegister + INT_ID_CHAN_STATUS) & CHAN_STATUS_FRAMING_ERROR))
	WRITEL(CHAN_STATUS_FRAMING_ERROR, ChP->pBaseRegister + INT_ID_CHAN_STATUS);
    return ret;
}

int uChanStatusParityError(CHANNEL_T *ChP)
{
    int ret;

    if ((ret = READL(ChP->pBaseRegister + INT_ID_CHAN_STATUS) & CHAN_STATUS_PARITY_ERROR))
	WRITEL(CHAN_STATUS_PARITY_ERROR, ChP->pBaseRegister + INT_ID_CHAN_STATUS);
    return ret;
}

int uRiActive(CHANNEL_T *ChP)
{
    return READL(ChP->pBaseRegister + INT_ID_CHAN_STATUS) & CHAN_STATUS_RI_ACTIVE;
}

int uCtsActive(CHANNEL_T *ChP)
{
    return READL(ChP->pBaseRegister + INT_ID_CHAN_STATUS) & CHAN_STATUS_CTS_ACTIVE;
}

int uDsrActive(CHANNEL_T *ChP)
{
    return READL(ChP->pBaseRegister + INT_ID_CHAN_STATUS) & CHAN_STATUS_DSR_ACTIVE;
}

int uCdActive(CHANNEL_T *ChP)
{
    return READL(ChP->pBaseRegister + INT_ID_CHAN_STATUS) & CHAN_STATUS_CD_ACTIVE;
}

int uXmitFifoEmpty(CHANNEL_T *ChP)
{
    return READL(ChP->pBaseRegister + INT_ID_CHAN_STATUS) &  CHAN_STATUS_XMIT_FIFO_EMPTY;
}

int uXmitShiftRegEmpty(CHANNEL_T *ChP)
{
    return READL(ChP->pBaseRegister + INT_ID_CHAN_STATUS) &  CHAN_STATUS_XMIT_SHIFT_REG_EMPTY;
}

int uDtrActive(CHANNEL_T *ChP)
{
    return READL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL) & SET_DTR_ACTIVE;
}

int uRtsActive(CHANNEL_T *ChP)
{
    return READL(ChP->pBaseRegister + INT_ID_CHAN_STATUS) & CHAN_STATUS_RTS_IN_STATUS;
}

int uRcvDataAvail(CHANNEL_T *ChP)
{
    return READL(ChP->pBaseRegister + INT_ID_CHAN_STATUS) &  CHAN_STATUS_RCV_DATA_AVAIL;
}

/*  
 *  Gets the bit in the Interrupt Status register corresponding to a channel (serial port).
 *  Inputs:   CtlP pointer to controller struct, port number
 *  Returns:  1 if set, 0 else
 */
int uGetInterruptStatus(CONTROLLER_T *CtlP, unsigned int port_number)
{
    if (port_number < 16) {
	return READL(CtlP->pBasePciMemory + CHANNEL_INT_STATUS) & Bit(port_number);
    }
    else {
	if (CtlP->has_2nd_asic)
	    return READL(CtlP->pBasePciMemory + 0x1000 + CHANNEL_INT_STATUS) & Bit(port_number - 16);
	else
	    return 0;
    }
}

/*  
 *  Enables/disables channel interrupts by clearing/setting the bit in the Interrupt Mask
 *  register corresponding to the channel.  
 *  Inputs:   ChP pointer to the channel struct
 */
void uEnChannelInterrupt(CHANNEL_T *ChP)
{
    if (ChP->channelNumber < 16) 
	SETBITW(ChP->pBaseRegister + CHANNEL_INT_MASK, Bit(ChP->channelNumber));
    else 
	SETBITW(ChP->pBaseRegister + CHANNEL_INT_MASK, Bit(ChP->channelNumber - 16));
}

void uDisChannelInterrupt(CHANNEL_T *ChP)
{
    if (ChP->channelNumber < 16) 
	CLEARBITW(ChP->pBaseRegister + CHANNEL_INT_MASK, Bit(ChP->channelNumber));
    else 
	CLEARBITW(ChP->pBaseRegister + CHANNEL_INT_MASK, Bit(ChP->channelNumber - 16));
}


/*  
 *  RCV_DATA_BYTE_STATUS Functions 
 *  Inputs:  The 16 bit byte|status read from RCV_DATA_BYTE_STATUS register (offset 0x0008) 
 *  Returns: 1 if set, else 0
 */
int uRcvByteParityError(unsigned short bytestatus)
{
    return bytestatus & RCV_BYTE_PARITY_ERROR;
}

int uRcvByteRcvOverrun(unsigned short bytestatus)
{
    return bytestatus & RCV_BYTE_RCV_OVERRUN;
}

int uRcvByteFramingError(unsigned short bytestatus)
{
    return bytestatus & RCV_BYTE_FRAMING_ERROR;
}

int uRcvByteBreakDetected(unsigned short bytestatus)
{
    return bytestatus & RCV_BYTE_BREAK_DETECTED;
}

int uRcvByteAddressByte(unsigned short bytestatus)
{
    return bytestatus & RCV_BYTE_ADDRESS_BYTE;
}

int uRcvByteTimeoutDetected(unsigned short bytestatus)
{
    return bytestatus & RCV_BYTE_TIMEOUT_DETECTED;
}

/*  
 *  Assert the modem/speaker reset signal on a port (all 8 reset signals if 
 *  port_number is 0xFF).  Calling program must range check the port_number 
 *  variable (0x00 - 0x07).
 */
int uModemReset(CONTROLLER_T *CtlP, unsigned short port_number)
{
//    WRITEW(Bit(port_number), CtlP->pBasePciMemory + MODEM_RESET);
    return 0;
}

int uSpeakerReset(CONTROLLER_T *CtlP, unsigned short port_number)
{
//    WRITEW(port_number, CtlP->pBasePciMemory + MODEM_RESET);
    return 0;
}

/*  
 *   Enables/disables the XANY function of the receive processor on a port. 
 */
void uEnXany(CHANNEL_T *ChP)
{
    WRITEB(URP_XANY_ENABLE, ChP->ucodeAddr + URP_XANY);
}

void uDisXany(CHANNEL_T *ChP)
{
    WRITEB(URP_XANY_DISABLE, ChP->ucodeAddr + URP_XANY);
}

/*  
 *   Enables/disables the receive processor transmitter on a port. 
 *   This is the Set XOFF and Set XON functions used in the Windows driver.  
 *   This puts the port in the XOFF or XON state, assuming software flow control is enabled.
 */
void uSetXon(CHANNEL_T *ChP)
{
    SETBITL(ChP->pBaseRegister + UART_CTRL, RCV_PROC_XMIT_EN);
}

void uSetXoff(CHANNEL_T *ChP)
{
    CLEARBITL(ChP->pBaseRegister + UART_CTRL, RCV_PROC_XMIT_EN);
}

/*  
 *   Returns the Transmitter XON/XOFF state when software flow control is active .
 *   Returns:  0 if XOFF'd, 1 if XON'd.
 */
int uXonXoffState(CHANNEL_T *ChP)
{
    return READL(ChP->pBaseRegister + UART_CTRL) & RCV_PROC_XMIT_EN;
}


/*  
 *   Enables/disables the One Stop Bit function on a port. 
 */
void uEnOneStopBit(CHANNEL_T *ChP)
{
    SETBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, ONE_STOP_BIT_EN);
}

void uDisOneStopBit(CHANNEL_T *ChP)
{
    CLEARBITL(ChP->pBaseRegister + XMIT_CTRL_RCV_CTRL, ONE_STOP_BIT_EN);
}

/*  
 *   Loads the channel's Receive Processor microcode.
 */
void uLoadUcode(CHANNEL_T *ChP)
{
    int i;

    for (i = 0; i < UCODESIZE; i++) {
	WRITEB(uCode[i], ChP->ucodeAddr + i);
    }
}

